import type { HighlighterCore, LanguageInput, ThemeRegistrationRaw } from '@shikijs/core'
import type { FunctionComponent, ReactNode } from 'react'
import type { ScrollViewProps, TextStyle, ViewStyle } from 'react-native'
import type { HighlighterStyleSheet } from './../../utils/style'
import { createHighlighterCore } from '@shikijs/core'
import { createJavaScriptRegexEngine } from '@shikijs/engine-javascript'
import React, { useEffect, useMemo, useState } from 'react'

import { ScrollView, StyleSheet, Text, View } from 'react-native'
import { trimNewlines } from 'trim-newlines'
import { getRNStylesFromShikiStyle } from './../../utils/style'

interface RendererNode {
  type: string
  value?: string
  children?: RendererNode[]
  properties?: {
    className?: string[]
  }
}

export interface SyntaxHighlighterProps {
  code: string
  language: string // Make sure this corresponds to a valid language input
  languageMap: Record<string, LanguageInput> // New prop for language map
  theme: ThemeRegistrationRaw
  textStyle?: TextStyle
  scrollViewProps?: ScrollViewProps
  containerStyle?: ViewStyle
  children?: ReactNode
}

const SyntaxHighlighter: FunctionComponent<SyntaxHighlighterProps> = ({ code, language, languageMap, theme, textStyle, scrollViewProps, containerStyle }) => {
  const [highlightedNodes, setHighlightedNodes] = useState<RendererNode[] | null>(null)

  useEffect(() => {
    const highlightCode = async () => {
      if (!languageMap[language]) {
        console.warn(`Warning: Language "${language}" not found in the language map. Falling back to "plaintext".`)
      }

      const highlighter: HighlighterCore = await createHighlighterCore({
        engine: createJavaScriptRegexEngine(),
        langs: Object.values(languageMap),
      })

      console.log('Highlighter:', highlighter)

      const tokens = highlighter.codeToHast(code, {
        lang: language,
        theme,
      })

      console.log('Tokens:', tokens || null)

      setHighlightedNodes(tokens.children || null)
    }

    highlightCode()
  }, [code, language, languageMap, theme])

  const stylesheet: HighlighterStyleSheet = useMemo(() => getRNStylesFromShikiStyle(theme), [theme])

  const getStylesForNode = (node: RendererNode): TextStyle[] => {
    const classes: string[] = node.properties?.className ?? []
    return classes.map((c: string) => stylesheet[c]).filter(Boolean) as TextStyle[]
  }

  const renderNode = (nodes: RendererNode[], keyPrefix = 'row'): ReactNode[] =>
    nodes.reduce<ReactNode[]>((acc, node, index) => {
      const keyPrefixWithIndex = `${keyPrefix}_${index}`
      if (node.children) {
        const styles = StyleSheet.flatten([textStyle, { color: stylesheet.shiki?.color }, ...getStylesForNode(node)])
        console.log('Styles:', styles)
        acc.push(
          <Text style={styles} key={keyPrefixWithIndex}>
            {renderNode(node.children, `${keyPrefixWithIndex}_child`)}
          </Text>,
        )
      }

      if (node.value) {
        acc.push(trimNewlines(String(node.value)))
      }

      return acc
    }, [])

  return (
    <ScrollView {...scrollViewProps} horizontal contentContainerStyle={[stylesheet.shiki, scrollViewProps?.contentContainerStyle, containerStyle]}>
      <View onStartShouldSetResponder={() => true}>{highlightedNodes && renderNode(highlightedNodes)}</View>
    </ScrollView>
  )
}

export default SyntaxHighlighter
